#!/usr/bin/env perl

# $Id: refgenomefasta2seqxml.pl 2944 2011-04-21 09:26:03Z dmessina $
# refgenomefasta2seqxml
# Copyright Dave Messina <dmessina@cpan.org>
# First version: 2010-01-14
#
# For questions, help, support, or suggestions:
# please email info@orthoxml.org
#
# For more information on SeqXML:
# please visit http://seqxml.org

use Modern::Perl;
use Bio::SeqIO;
use File::Basename;
use File::Spec;
use Getopt::Long;
use Digest::MD5 qw(md5_hex);

my ( $source, $sourceVersion, $strip_extra, $species_override, $taxID_override,
     $md5, $debug );
GetOptions(
    "source=s"        => \$source,
    "sourceVersion=s" => \$sourceVersion,
    "stripextra"      => \$strip_extra,
    "species=s"       => \$species_override,
    "taxid=i"         => \$taxID_override,
    "md5"             => \$md5,
    "debug"           => \$debug,
);

my $usage = "
refgenomefasta2seqxml - converts FASTA files from the Quest for Orthologs
                        Primary Protein Sets into SeqXML
                        
Usage: refgenomefasta2seqxml ref1.fasta ref2.fasta ... refN.fasta

Options:
--source        <string>  set the source (or database) name in the header
--sourceVersion <string>  set the source (or database) version in the header
--stripextra              strip extra bits from filename if possible
--species       <string>  manually specify species string
                          (overrides genus and species in filename)
--taxid         <number>  manually specify species taxID
                          (overrides taxID in filename)
--md5                     include the md5 checksum hexadecimal digest of
                          the sequence as a property.
--debug                   turn on debugging messages

This program expects files that are named like NCBItaxID_genus_species.fasta.
For example, 9606_homo_sapiens.fasta. The filename can have extra bits also,
as long as they are separated by a '_' and after the species name. For
example, these are legal:

9606_homo_sapiens_collected_by_Bob.fasta
9606_homo_sapiens_2010-01-01.fasta
9606_homo_sapiens_and_a_partridge_in_a_pear_tree.fasta

This species info will be included in each record in the species tag.

This program expects FASTA headers that look like this:
>UniProtKB/Swiss-Prot:Q3KM44 Gene:CTA_0340 GN:rplA PE:3 Description:50S ribosomal protein L1

Namely, key-value pairs separated by colons. The first field is taken as the
sequence ID, and if it's a colon-separated key-value pair, the part after the
colon is assigned as the ID. e.g. UniProtKB/Swiss-Prot:Q3KM44 becomes Q3MK44

Subsequent fields are processed as follows:

Tag                     stored in SeqXML as:
---                     --------------------
Description             description
CGD                     DBRef of type 'gene'
dictyBase               DBRef of type 'gene'
EntrezGene              DBRef of type 'gene'
GeneDB_Pfalciparum      DBRef of type 'gene'
GeneDB_Spombe           DBRef of type 'gene'
EcoCyc                  DBRef of type 'protein'
EMBL                    DBRef of type 'mRNA'
ENSEMBL                 DBRef of type 'gene', 'protein', 'mRNA' *
FB                      DBRef of type 'gene'
HGNC                    DBRef of type 'gene'
IPI                     DBRef of type 'protein'
MGI                     DBRef of type 'gene'
protein_id              DBRef of type 'protein'
RefSeq_dna              DBRef of type 'mRNA'
RefSeq_peptide          DBRef of type 'protein'
RGD                     DBRef of type 'gene'
SGD                     DBRef of type 'gene'
TAIR                    DBRef of type 'gene'
UniProtKB               DBRef of type 'protein'
UniProtKB/Swiss-Prot    DBRef of type 'protein'
UniProtKB/TrEMBL        DBRef of type 'protein'
WB                      DBRef of type 'gene'
ZFIN                    DBRef of type 'gene'
Gene                    DBRef of type 'gene'
gene                    DBRef of type 'protein' **
GN                      DBRef of type 'gene'
PE                      property

All others will be stored as properties.

* Note that ENSEMBL is a special case, and we attempt to decode what type of
  ENSEMBL identifier is used and assign the correct type.
  
** In the Arabidopsis thaliana reference genome fasta (as of 11/14/10),
   Uniprot protein IDs are marked as 'gene'.

For more details on the Quest for Orthologs Primary Protein Sets project, see:
    http://www.ebi.ac.uk/reference_proteomes/

For more details on SeqXML, see http://seqxml.org
";
die $usage unless @ARGV;

# known tag lookup
my %tags = (
    'Description'          => ['description'],
    'CGD'                  => ['DBRef', 'gene'],
    'dictyBase'            => ['DBRef', 'gene'],
    'EntrezGene'           => ['DBRef', 'gene'],
    'EcoCyc'               => ['DBRef', 'protein'],
    'EMBL'                 => ['DBRef', 'mRNA'],
    'ENSEMBL'              => ['ENSEMBL'],
    'FB'                   => ['DBRef', 'gene'],
    'GeneDB_Pfalciparum'   => ['DBRef', 'gene'],
    'GeneDB_Spombe'        => ['DBRef', 'gene'],
    'HGNC'                 => ['DBRef', 'gene'],
    'IPI'                  => ['DBRef', 'protein'],
    'MGI'                  => ['DBRef', 'gene'],
    'protein_id'           => ['DBRef', 'protein'],
    'RefSeq_dna'           => ['DBRef', 'mRNA'],
    'RefSeq_peptide'       => ['DBRef', 'protein'],
    'RGD'                  => ['DBRef', 'gene'],
    'SGD'                  => ['DBRef', 'gene'],
    'TAIR'                 => ['DBRef', 'gene'],
    'UniProtKB/Swiss-Prot' => ['DBRef', 'protein'],
    'UniProtKB'            => ['DBRef', 'protein'],
    'UniProtKB/TrEMBL'     => ['DBRef', 'protein'],
    'WB'                   => ['DBRef', 'gene'],
    'ZFIN'                 => ['DBRef', 'gene'],
    'Gene'                 => ['DBRef', 'gene'],
    'gene'                 => ['DBRef', 'protein'],
    'GN'                   => ['DBRef', 'gene'],
    'PE'                   => ['property'],
);

# maps a SeqXML tag type to the subroutine which handles creating
# that type of record in the sequence object
my %codemap = (
    'description' => \&description,
    'DBRef'       => \&DBRef,
    'ENSEMBL'     => \&ENSEMBL,
    'property'    => \&property,
);

foreach my $infile (@ARGV) {

    # parse filename
    my ( $base, $path, $suffix ) = fileparse( $infile, qr/\.[^.]*/ );
    if ($debug) { say STDERR $base; }

    # parse species info from filename (and upcase Genus)
    my ( $taxID, $genus, $species, $rest ) = split( '_', $base );
    unless ( ( $taxID =~ /[0-9]{1}/ ) && defined($genus) && defined($species) ) {
        print STDERR "couldn't parse filename $infile."
          . " Won't use it for species information.\n";
        undef($taxID);
        undef($genus);
        undef($species);
    }
    $genus = ucfirst($genus);
    
    # override species and taxID if passed in as parameters
    # otherwise use that from filename
    my $species_string;
    if (defined($species_override)) { $species_string = $species_override; }
    else { $species_string = join(' ', $genus, $species); }
    if ( defined($taxID_override) ) { $taxID = $taxID_override; }

    # input stream
    my $in = Bio::SeqIO->new(
        '-format'   => 'fasta',
        '-file'     => $infile,
        '-alphabet' => 'protein',
    );

    # output filename will be NCBItaxID_genus_species.xml if --stripextra
    # otherwise it'll be same as input filename but .xml suffix
    my $newbase;
    if ($strip_extra) {
        if ( defined($taxID) && defined($species_string) ) {
            $newbase = join( '_', $taxID, $species_string );
        }
        else {
            print STDERR
              "couldn't get taxID and species string, so not stripping extra\n";
            $newbase = $base . '.xml';
        }
    }
    else { $newbase = $base . '.xml'; }
    my $outfile = File::Spec->catfile( $path, $newbase );

    # create output stream
    my $out;
    if ( $source || $sourceVersion ) {

        # fake values to appease my mistake of requiring both
        $source        = '' unless defined $source;
        $sourceVersion = '' unless defined $sourceVersion;

        $out = Bio::SeqIO->new(
            '-format'        => 'seqxml',
            '-file'          => ">$outfile",
            '-alphabet'      => 'protein',
            '-source'        => $source,
            '-sourceVersion' => $sourceVersion,
        );

    }
    else {
        $out = Bio::SeqIO->new(
            '-format'   => 'seqxml',
            '-file'     => ">$outfile",
            '-alphabet' => 'protein',
        );
    }

    # read each entry in the input file
    # parse the header
    # assign each part of the header to the right seqXML tag
    # write the output
    while ( my $seq_obj = $in->next_seq ) {
        
        # convert UniProt:A12345 ID into just A12345
        my ($db, $id) = split ':', $seq_obj->display_id(), 2;
        if (defined $db && defined $id) {
            $seq_obj->display_id($id);
            my $annotation_obj = Bio::Annotation::SimpleValue->new(
                '-tagname' => 'source',
                '-value'   => $db,
            );
            $seq_obj->add_Annotation($annotation_obj);
        }

        # parse the rest of the FASTA header
        if ( $seq_obj->desc ) {
            my $tag_hash = parse_desc($seq_obj);

            # assign tags to appropriate part of seq obj
            process_tags( $seq_obj, $tag_hash );
        }

        # calculate MD5 checksum and add to seq obj
        if ($md5) {
            my $chksum = md5_hex( $seq_obj->seq );
            unless ($chksum) {
                print STDERR ( "couldn't get md5 for ", $seq_obj->id, "!\n" );
            }
            my $hashref = { $chksum => 1 };
            property( $seq_obj, 'md5', $hashref );
        }

        # set species info
        # WARNING: This overwrites existing species info
        # if it's there from the input sequence record
        if ( defined $taxID && defined $species_string ) {
            my $species_obj = Bio::Species->new( -ncbi_taxid => $taxID );
            $species_obj->node_name($species_string);
            $seq_obj->species($species_obj);
        }

        $out->write_seq($seq_obj);
    }
}

# Name: parse_desc
# Desc: extracts colon-joined tag-value pairs from
#       the Seq object's description,
#       stores the unique values in a tag-keyed hash
# Args: a Bio::Seq object
# Returns: a hashref with tags as keys and a hash of tag values as values
sub parse_desc {
    my ($seq_obj) = @_;

    my $desc = $seq_obj->desc;

    # wipe out existing description so that if there isn't a tagged
    # Description:foo description, we don't accidentally have the
    # old description string with all the tag:value pairs in it
    $seq_obj->desc(undef);

    # for storing all the data. This is the hashref that gets returned.
    my $parts = {};

    # grab free text at the beginning of the string,
    # (it's the description)
    # or the first key:value pair
    my @tag_value_pairs = $desc =~ /(.+?)\s(?=\S+?:.+)/p;
    
    # if there's only one blob in desc, above will fail
    if ( scalar @tag_value_pairs == 0 ) {
        @tag_value_pairs = ($desc);
    }
    else {
        # loop through all tag:value pairs
        # except for the last one.
        # why? because by using a look-ahead to the next pair,
        # this pattern will properly capture values that have
        # spaces, like:
        # Test: a multi-word pattern
        my $remaining = ${^POSTMATCH};
        if ($remaining) {
            my @pairs = $remaining =~ /(\S+?:.+?)\s(?=\S+?:)/g;
            push @tag_value_pairs, @pairs;
    
            # grab the last pair
            my $lastbit = ${^POSTMATCH};
            if ($lastbit) {
    
                my @lastpair = $lastbit =~ /(\S+?:.+)$/;
    
                # strip trailing spaces
                $lastpair[0] =~ s/\s+$//g;
    
                push @tag_value_pairs, @lastpair;
            }
        }
    }

    # split all the pairs into tags and values,
    # store the values in a hash of hashes indexed by tag name
    # note that this allows multiple values for the same tag
    # (for when we have for example multiple gene IDs)
    for (my $i = 0; $i < @tag_value_pairs; $i++) {
        my $pair = $tag_value_pairs[$i];
        
        # only split on the first ':'; subsequent ones are ignored
        if ( $pair =~ /:/ ) {
            my ( $tag, $value ) = split( /:/, $pair, 2 );

            # if pairs after Description don't have known tags, join them
            # into Description's value
            if ($tag eq 'Description' && $i != $#tag_value_pairs) {
                my @rest = splice @tag_value_pairs, $i+1;
                my ( $rest_tag, $rest_value ) = split( /:/, $rest[0] );
                if ( !exists $tags{$rest_tag} ) {
                    my $rest = join '', @rest;
                    $value = $value . ' ' . $rest;
                    if ($debug) { say STDERR "concatenated Description: $value"; }
                }
            }

            # check for Description values that contained a ':', didn't get
            # caught by join code above, and got split
            if ( $tag eq 'Description' ) {
                if ( $value =~ /similar to\s*$/i ) {
                    print STDERR " - possibly broken desc in ",
                      $seq_obj->primary_id(),
                      ": $value\n";
                }
            }

            $parts->{$tag}->{$value} = 1;
        }
        else {
            warn(
                "malformed part of description in ",
                $seq_obj->primary_id(),
                ". Not a tag-value pair!: $pair\n"
            );
        }
    }
    return $parts;
}

# Name: process_tags
# Desc: reads tag-value pairs from a hash
#       looks up in %tags which type of tag goes where in a SeqXML object
#       creates the right type of Annotation object for it
#       and adds it to the seq object
# Args: a Bio::Seq object, a hash of hashes (produced by parse_desc())
# Returns: nothing; modifies the seq object
sub process_tags {
    my ( $seq_obj, $tag_hash ) = @_;

    foreach my $tagname ( keys %$tag_hash ) {
        my $tagvalues = $$tag_hash{$tagname};

        if ( $tags{$tagname} ) {
            my $sub = $tags{$tagname}->[0];
            $codemap{$sub}->( $seq_obj, $tagname, $tagvalues );
        }
        else {
            property( $seq_obj, $tagname, $tagvalues );

            if ($debug) {
                print STDERR "WARNING: unknown tag $tagname in ",
                  $seq_obj->id, ". Adding it as a property.\n";
            }
        }
    }
}

sub description {
    my ( $seq_obj, $tagname, $tagvalues ) = @_;

    if ( scalar keys %$tagvalues > 1 ) {
        die( "more than one description in ", $seq_obj->id, "\n" );
    }
    else {
        my ($desc) = keys %$tagvalues;
        $seq_obj->desc($desc);
    }
}

sub DBRef {
    my ( $seq_obj, $tagname, $tagvalues ) = @_;

    # If there is a type specified
    if (defined $tags{$tagname}->[1]) {
        my $DBRef_type = $tags{$tagname}->[1];
        
        foreach my $tagvalue ( keys %$tagvalues ) {
            my $annotation_obj = Bio::Annotation::DBLink->new(
                -primary_id => $tagvalue,
                -database   => $tagname,
                -type       => $DBRef_type,
                -tagname    => 'dblink',
            );
            $seq_obj->add_Annotation($annotation_obj);
        }        
    }
    else {
        die $seq_obj->display_id, ": DBRefs must have a type.\n";
    }
}

# ENSEMBL gets a special type because, depending on the ID, we'll make it
# a different type of DBRef
sub ENSEMBL {
    my ( $seq_obj, $tagname, $tagvalues ) = @_;
    
    foreach my $tagvalue ( keys %$tagvalues ) {
        # Some ENSEMBL IDs are like 'ENSMUSG00000' or ENSERNOP0000'
        # The 7th letter of the ID is G, P, or T, indicating
        # gene, protein, or mRNA
        my $seventh_letter = substr $tagvalue, 6, 1;
        my $DBRef_type;
        given ($seventh_letter) {
            when ($seventh_letter eq 'G') {
                $DBRef_type = 'gene';
            }
            when ($seventh_letter eq 'P') {
                $DBRef_type = 'protein';                
            }
            when ($seventh_letter eq 'T') {
                $DBRef_type = 'mRNA';                
            }
            when ($seventh_letter =~ /[0-9]/) {
                # other ENSEMBL IDs are like 'ENSG0000', 'ENSP0000', and
                # 'ENST0000'. This section handles those
                my $fourth_letter = substr $tagvalue, 3, 1;
                given ($fourth_letter) {
                    when ($fourth_letter eq 'G') {
                        $DBRef_type = 'gene';
                    }
                    when ($fourth_letter eq 'P') {
                        $DBRef_type = 'protein';                
                    }
                    when ($fourth_letter eq 'T') {
                        $DBRef_type = 'mRNA';                
                    }
                    default {
                        die "unexpected ENSEMBL ID: $tagvalue\n";
                    }
                }
            }
            default {
                die "unexpected ENSEMBL ID: $tagvalue\n";
            }
        }
        
        my $annotation_obj = Bio::Annotation::DBLink->new(
            -primary_id => $tagvalue,
            -database   => 'ENSEMBL',
            -type       => $DBRef_type,
            -tagname    => 'dblink',
        );
        $seq_obj->add_Annotation($annotation_obj);
    }
}

sub property {
    my ( $seq_obj, $tagname, $tagvalues ) = @_;

    foreach my $tagvalue ( keys %$tagvalues ) {
        my $annotation_obj = Bio::Annotation::SimpleValue->new(
            -tagname => $tagname,
            -value   => $tagvalue,
        );
        $seq_obj->add_Annotation($annotation_obj);
    }

}
